unit ICText;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements a simple wrapper around the system's TextEdit routines.
	The module is very simple, it's main function is to regularise the interface
	to TextEdit, which can be a bit wacky at times.

	The external interface for this module deals entirely in textRef's, a universal
	pointer type.  This is basically a cookie that the caller gets using a TextCreate
	and must pass in to all the other routines.  Internally this translates to a
	TextItemPtr, so the first thing most routines do is cast the textRef to a
	TextItemPtr and then continue on from there.
*)

interface

	uses
		Dialogs;

	(* ***** Create and Destroy ***** *)
		
	function TextCreate (var textRef: univ Ptr; window: DialogPtr; item: integer; 
											font: integer; size: integer; locked: Boolean): OSStatus;
		(* Creates a text item associated with the specified dialog item.  The
			font and size parameters determine the font used to display the text.
			The locked parameters determines whether the text should display an
			insertion point or not.
		*)
	
	procedure TextDestroy (var textRef: univ Ptr);

	(* ***** Basic Event Handling ***** *)
		
	(* Call these routines in response to basic system events. *)
	
	procedure TextActivate (textRef: univ Ptr; activate: Boolean);
	procedure TextClick (textRef: univ Ptr; event: EventRecord);
	procedure TextIdle (textRef: univ Ptr);
	procedure TextKey (textRef: univ Ptr; event: EventRecord);
	procedure TextDraw (textRef: univ Ptr);
	
	(* ***** Selection Handling ***** *)
	
	(* This routines allow you to get and set the text selection. *)
	
	procedure TextGetSelect (textRef: univ Ptr; var selStart, selEnd: integer);
	procedure TextSetSelect (textRef: univ Ptr; selStart, selEnd: integer);

	(* ***** Data Handling ***** *)
		
	procedure TextGetSize (textRef: univ Ptr; var textSize: longint);
		(* Sets textSize to the number of characters in the text item. *)
	
	procedure TextInsert (textRef: univ Ptr; textH: Handle);
		(* Inserts the characters in textH into the text at the current insertion point. It makes
			a copy of the data, so the caller must still dispose of the handle.
		*)
	
	procedure TextGet (textRef: univ Ptr; textH: Handle);
		(* Sets the contents of textH to be the contents of the text item.  The caller
			is responsible for creating textH, this routine resizes it and copies
			the data in, and then the caller is responsible for disposing it.
		*)

	(* ***** Miscellaneous Stuff ***** *)
	
	procedure TextMove (textRef: univ Ptr; newTextItemRect: Rect);
		(* Moves the text item to the new position. *)

	(* ***** Edit Menu Support ***** *)

	(* Call these routines in response to choices from the Edit menu. *)
		
	procedure TextCut (textRef: univ Ptr);
	procedure TextCopy (textRef: univ Ptr);
	procedure TextPaste (textRef: univ Ptr);
	procedure TextClear (textRef: univ Ptr);

	(* ***** Initialisation ***** *)
	
	procedure InitICText;
	
implementation

	uses
		ICDialogs, ICMiscSubs, Scrap;

	var
		gNopCaretHookUPP : CaretHookUPP;

{$ifc GENERATINGCFM}

	procedure NopCaretHook(var r: Rect; pTE: TEPtr);
	begin
		{$unused r}
		{$unused pTE}
		(* do nothing -- Hey, that's why it's called NopCaretHook! *)
	end; (* NopCaretHook *)

{$elsec}

	procedure NopCaretHook(var r: Rect; pTE: TEPtr);
		asm;
	begin
		move.l			(a7)+,a0			;		(* junk pointer to Rect on stack *)
		rts												(* return *)
	end; (* NopCaretHook *)

{$endc}

	(* The TextItemRecord is the primary internal data structure for this module.
		It's basically a wrapper around a standard system TEHandle, along with a few
		other extra fields to help keep track of where to draw the text.
	*)
	type
		TextItemRecord =
			record
				window: DialogPtr;
				item: integer;
				textEditH: TEHandle;
			end;
		TextItemPtr = ^TextItemRecord;

	(* ***** Create and Destroy ***** *)

	function TextCreate (var textRef: univ Ptr; window: DialogPtr; item: integer; 
											font: integer; size: integer; locked: Boolean): OSStatus;
		var
			err: OSStatus;
			textItem: TextItemPtr;
			savedWindowState: SavedWindowInfo;
			viewRect : Rect;
			destRect: Rect;
			textFontInfo : FontInfo;
			lineHeight : integer;
	begin
		textRef := NewPtr(sizeof(TextItemRecord));
		err := MemError;
		if err = noErr then begin
		
			(* Fill out the various fields of the text item. *)
			textItem := TextItemPtr(textRef);
			textItem^.window := window;
			textItem^.item := item;
			textItem^.textEditH := nil;
			
			(* Enter the window and calculate the view and dest rectangles. *)
			EnterWindow(window, font, size, [], savedWindowState);
			GetDItemRect(window, item, destRect);
			viewRect := destRect;
			GetFontInfo(textFontInfo);
			lineHeight := textFontInfo.leading + textFontInfo.ascent + textFontInfo.descent;
			destRect.bottom := destRect.top + (destRect.bottom - destRect.top) div lineHeight * lineHeight;
			
			(* Create the TEHandle and setup a few parameters in it. *)
			textItem^.textEditH := TENew(destRect, viewRect);
			if locked then begin
				(* Disable the insertion point. *)
				textItem^.textEditH^^.caretHook := gNopCaretHookUPP;
			end; (* if *)
			TEAutoView(true, textItem^.textEditH);
			
			(* Clean up. *)
			ExitWindow(savedWindowState);
			if err <> noErr then begin
				TextDestroy(textRef);
			end; (* if *)
		end; (* if *)
		TextCreate := err;
	end; (* TextCreate *)

	procedure TextDestroy (var textRef: univ Ptr);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if textRef <> nil then begin
			if textItem^.textEditH <> nil then begin
				TEDispose(textItem^.textEditH);
			end; (* if *)
			DisposePtr(textRef);
			textRef := nil;
		end; (* if *)
	end; (* TextDestroy*)

	(* ***** Basic Event Handling ***** *)

	procedure TextActivate (textRef: univ Ptr; activate: Boolean);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if activate then begin
			TEActivate(textItem^.textEditH);
		end else begin
			TEDeactivate(textItem^.textEditH);
		end; (* if *)
	end; (* TextActivate *)

	procedure TextClick (textRef: univ Ptr; event: EventRecord);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		SetPort(textItem^.window);
		GlobalToLocal(event.where);
		if PtInRect(event.where, textItem^.textEditH^^.viewRect) then begin
			TEClick(event.where, band(event.modifiers, shiftKey) <> 0, textItem^.textEditH);
		end; (* if *)
	end; (* TextClick *)

	procedure TextIdle (textRef: univ Ptr);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		TEIdle(textItem^.textEditH);
	end; (* TextIdle*)

	procedure TextKey (textRef: univ Ptr; event: EventRecord);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if band(event.modifiers, cmdKey) = 0 then begin
			TEKey(chr(band(event.message, charCodeMask)), textItem^.textEditH);
		end; (* if *)
	end; (* TextKey *)

	procedure TextDraw (textRef: univ Ptr);
		var
			textItem: TextItemPtr;
			itemRect: Rect;
	begin
		textItem := TextItemPtr(textRef);
		GetDItemRect(textItem^.window, textItem^.item, itemRect);
		EraseRect(itemRect);
		TEUpdate(textItem^.textEditH^^.viewRect, textItem^.textEditH);
	end; (* TextDraw *)

	(* ***** Selection Handling ***** *)

	procedure TextGetSelect (textRef: univ Ptr; var selStart, selEnd: integer);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		selStart := textItem^.textEditH^^.selStart;
		selEnd := textItem^.textEditH^^.selEnd;
	end; (* TextGetSelect *)

	procedure TextSetSelect (textRef: univ Ptr; selStart, selEnd: integer);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		TESetSelect(selStart, selEnd, textItem^.textEditH);
	end; (* TextSetSelect *)

	(* ***** Data Handling ***** *)

	procedure TextGetSize (textRef: univ Ptr; var textSize: longint);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		textSize := GetHandleSize(textItem^.textEditH^^.hText);
	end; (* TextGetSize *)

	procedure TextInsert (textRef: univ Ptr; textH: Handle);
		var
			textItem: TextItemPtr;
			s: SInt8;
	begin
		textItem := TextItemPtr(textRef);
		s := HGetState(textH);
		HLock(textH);
		TEInsert(textH^, GetHandleSize(textH), textItem^.textEditH);
		HSetState(textH, s);
	end; (* TextInsert *)

	procedure TextGet (textRef: univ Ptr; textH: Handle);
		var
			textItem: TextItemPtr;
			textInTextItem: Handle;
			sizeOfTextInTextItem: longint;
	begin
		textItem := TextItemPtr(textRef);
		textInTextItem := Handle(TEGetText(textItem^.textEditH));
		sizeOfTextInTextItem := GetHandleSize(textInTextItem);
		SetHandleSize(textH, sizeOfTextInTextItem);
		if MemError = noErr then begin
			BlockMoveData(textInTextItem^, textH^, sizeOfTextInTextItem);
		end else begin
			SetHandleSize(textH, 0);
		end; (* if *)
	end; (* TextGet *)

	(* ***** Miscellaneous Stuff ***** *)

	procedure TextMove (textRef: univ Ptr; newTextItemRect: Rect);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		textItem^.textEditH^^.viewRect := newTextItemRect;
		textItem^.textEditH^^.destRect := newTextItemRect;
		TECalText(textItem^.textEditH);
	end; (* TextMove *)

	(* ***** Edit Menu Support ***** *)

	procedure TextCut (textRef: univ Ptr);
		var
			textItem: TextItemPtr;
			junk: longint;
	begin
		textItem := TextItemPtr(textRef);
		TECut(textItem^.textEditH);
		junk := ZeroScrap;
		junk := TEToScrap;
	end; (* TextCut *)

	procedure TextCopy (textRef: univ Ptr);
		var
			textItem: TextItemPtr;
			junk: longint;
	begin
		textItem := TextItemPtr(textRef);
		TECopy(textItem^.textEditH);
		junk := ZeroScrap;
		junk := TEToScrap;
	end; (* TextCopy *)

	procedure TextPaste (textRef: univ Ptr);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if TEFromScrap = noErr then begin
			TEPaste(textItem^.textEditH);
		end; (* if *)
	end; (* TextPaste *)

	procedure TextClear (textRef: univ Ptr);
		var
			textItem: TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		TEDelete(textItem^.textEditH);
	end; (* TextClear *)

	procedure InitICText;
	begin
		gNopCaretHookUPP := NewCaretHookProc(@NopCaretHook)
	end; (* InitICText *)

end. (* ICText *)
