unit ICWindowGlobals;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module is the core of the Internet Config application's window management
	system. It exports all of the data structures that represent the state of all
	the windows and a number of routines for the use of Whats.
	
	A "What" is the term we made up to describe an extended dialog item.
	IC uses modeless dialogs throughout, but dialog items have insufficent
	expressive power, so we keep a parallel data structure that maintains
	extra information. This data structure is made up of Whats, one for
	each active dialog item in the dialog.
	
	Why "What"?  Well, we were sitting in ResEdit saying "What on earth
	are we going to call these things?" and eventually decided that "What"s
	would do (-:
	
	Each What contains a pointer to its class.  The class is data structure
	used to hold a bunch of ProcPtrs that control the behaviour of the What.
	There is a class structure for each common type of dialog item,
	for example text, checkboxes, and popup menus.
	
	The Whats are created when the window is opened. For each DLOG
	resource, there is a corresponding What Template resource that holds
	the extra information needed to create the Whats for the items
	in the dialog. The most important piece of data in the What Template
	resource is the TypeID, which is used to find the class that controls
	the What, for example the TypeID would be 'TEXT' to create a text
	item.
	
	Because the Whats are all accessed through ProcPtrs, and the
	placement of the Whats within a dialog is controlled by the DITL
	and the WHAT resources, there is a high degree of separation
	between the Internet Config application and the data it displays.
	In fact we've tried to maintain this separation as much as
	possible, although sometimes the layers break down. The two
	key areas of layer breaking are the Text Whats, which have
	all sorts of weird interactions with the window management
	layer, and the Edit Preference AppleEvent, which has to talk
	to the structured whats (File Mapping and Helpers) to select
	the right element.  Maybe one day I'll clean this up, but
	at the moment I'm prepared to accept a little layer breaking
	for a lot of convenience.

	Part of the layered design is that the Whats (generally)
	don't need to access any modules above this one. This module
	exports all the data structures and routines that the Whats
	need. They don't need to access ICWindows, ICEvents, 
	ICDocuments, and so on.  Again, this breaks down in the case
	of the Button Whats, but it's a good rule of thumb.

	Some common gotchas with this design:
	
	1. gWhatClasses is a fixed sized array.  When you add a class,
		you have to remember to increase kMaxClass.
	
	2. gWindows[wt].items is a fixed sized array. The upper
		bound is kMaxItemsPerWindow.  If you create a window
		with more active items in it, you must remember to increase
		this number.
		
		Note that this isn't the limit of the number of dialog
		items, it's just the limit of the number of What
		items.  You can have extra dialog items beyond this limit,
		as long as these items don't need special behaviour
		beyond that supplied by the Dialog Manager.

	3.	Any element in the gWindows[wt].items array may
		be nil.  You have to test it each time.  This radically cuts
		down on the memory usage for windows that contain a lot
		of items that don't require Whats (like static text items)
		but it does complicate programming a bit.
		
		Note that a What may safely assume that its own element
		pointer is not nil, otherwise it wouldn't have been called.
		
	4. The What's closeProc can be called without the openProc
		being called.  This is due to the way the opening and closing
		routines work.  A What's closeProc should guard against
		this, using the knowledge that the instance data is
		initialised to zero before either the openProc or closeProc 
		is called.
	
	5.	Complex Whats, like the one that drives the File Mapping 
		dialog, can often have other Whats that help them along.
		For example, the File Mapping What has a corresponding
		File Button What that whose sole purpose is to provide
		a way to route the clicks on the buttons to the File
		Mapping What as a whole.
		
		This leads so interactions between Whats, some subtle and
		some not so subtle. The File Mapping What is not very
		subtle, and hard codes the dialog item numbers of the
		various components of the dialog. On the other hand,
		the RGBColour What has a very subtle connection to the
		user item that draws the sample of the colour, namely
		that the user item must have a dialog item number one
		less than the RGBColour What itself.
		
		Note that normally there is a central What that coordinates
		the loading and saving of the What's data, and then there
		are ancilliary Whats that collect user actions.
*)

interface

	uses
		Files,
		WIndows,
		Dialogs,

		InternetConfig;
	
	(* ***** Basic Declarations ***** *)
	
	(* IC is in the fortunate position of knowing exactly how many windows it
		might have open at any point in time. So rather than mess around with
		dynamic allocation, it can simply statically allocate the windows
		it needs in a global array. This global array is called gWindows (see
		the description below), and it's indexed by WindowType.  WindowType
		is one of the core data type, because it can be used to denote any
		window in the application.
		
		Windows are divided up into "Information" windows (those that show
		IC preference information) and "other" windows. The about box is
		the obvious example of an other window, but also the Main window
		is technically not an information window. The key distinction between
		these two types of windows is in ICWindows.WindowsFlushAll and
		ICWindows.WindowsCloseAll, which flush and close information
		windows (ie those showing preferences) and not other windows.
	*)
	type
		WindowType = (WT_None, WT_About, WT_Main,
			WT_Personal, WT_Email, WT_News, WT_Web, WT_FileTransfer,
			WT_OtherServices, WT_Font, WT_FileMapping, WT_Helper, WT_Firewalls,
			WT_Last);

	const
		kMaxItemsPerWindow = 15;
		(* The maximum number of items in a window. Like the Dialog Manager,
			the items are indexed from 1.
		*)

		kMaxClass = 16;
		(* The maximum number of what classes. What classes are
			held in a global array gWhatClasses whose size is defined by
			this.
		*)

	(* ***** What Template ***** *)
	
	(* A What Template is a resource (type WHAT) that's used to construct the
		Whats for a window when the window is opened. The WHAT resource has
		the same resource ID as the DLOG resource that it's paired with.
		Items in the template are paired with items in the DITL by item number.
		The template provides extra information that we need to run the dialog
		item associated with the template item. This includes:
		
		1. The type ID of the class that controls the item.
		2. The IC key of the preference that this item controls.
		3. Flags that are interpreted by the class.
		
		For some classes of Whats (eg About Whats) the key is ignored.
		
		Because WhatTemplateRecord maps directly on to a resource,
		we have to enforce 68K alignment for it.
	*)

{$PUSH}
{$ALIGN MAC68K}

	type
		WhatTemplateRecord =
			record
				templateKey: Str31;
				templateTypeID: OSType;
				templateFlags: longint;
			end;
		WhatTemplateArray = array[1..kMaxItemsPerWindow] of WhatTemplateRecord;
		WhatTemplateArrayPtr = ^WhatTemplateArray;
		WhatTemplateArrayHandle = ^WhatTemplateArrayPtr;

{$ALIGN RESET}
{$POP}

	(* ***** What Class ***** *)

	(* The What Class encapsulates the code that controls the What.
		Because many Whats have similar behaviour, the class is abstracted
		out from the What and shared between many Whats. You can
		think of the Whats as an instance of the What Class.
		
		There are three important components of a class. The first is the 
		class type ID, which is used to match classes up with whats
		as they are created from the What Template. 
		
		The second is a bunch of ProcPtrs that are used to control the
		interaction between IC and the Whats.  Any of these ProcPtrs may
		be nil, in which case it is assumed that the class is not interested
		in the corresponding event.
		
		The third component is the cursorID field, which contains the
		resource ID of the CRSR that the What wants the cursor set
		to when it is over the What. For example, the Text What sets
		this field to iBeam.  The What can override this decision by setting
		up a cursorProc. 
	*)
	
	type
		WhatInitProcPtr = function(dummy : integer) : OSStatus;
		(* The initProc is called once at the start of the execution of the
			application.  What classes use it to create any data that's global
			across all instances of the class, normally UPP.
		   The dummy parameter exists to avoid confusion with Metrowerks Pascal,
		   which has a hard time telling the difference between a pointer
		   to a function that takes no parameters, and the result of calling
		   a pointer to a function that takes no parameters.
		*)

		WhatOpenProcPtr = function (wt: WindowType; item: integer): OSStatus;
		(* The openProc is called when the window containing the What
			is created. The wt and item parameters describe the window
			and item number of the What being opened. The current port will be
			the window associated with wt. If the What is in an informational
			window, it may rely on being inside an ICBegin/ICEnd pair.
			
			If the What returns an error, the error will prevent the window opening.
			The What should still be prepared for its closeProc to be called.
		*)
		
		WhatFlushProcPtr = function (wt: WindowType; item: integer): OSStatus;
		(* The flushProc is called when the What is saved.  The What should
			keep track of whether it has been modified or not, and only attempt
			to write out its preference if it has been modified.
			
			The wt and item parameters describe the window and item number
			of the What.  The current port will be the window associated with wt.
			If the What is in an informational window, it may rely on being inside 
			an ICBegin/ICEnd pair.
			
			If the What returns an error, the error will be displayed to the user.
		*)

		WhatCloseProcPtr = function (wt: WindowType; item: integer): OSStatus;
		(* The closeProc is called when the What is destroyed.  The closeProc
			may be called even if the openProc returns an error, or even if the
			openProc hasn't been called. The only guarantee is that the instance
			data in the WhatRecord will be cleared before the openProc or
			closeProc is called.  The closeProc should be very defensive.
			
			The wt and item parameters describe the window and item number
			of the What.  The current port will be the window associated with wt.
			If the What is in an informational window, it may rely on being inside 
			an ICBegin/ICEnd pair.
			
			If the What returns an error, the error will be displayed to the user.
		*)

		WhatKeyProcPtr = function (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* The keyProc is called when the user types a key in the window
			associated with wt and item, and this What is the one denoted by the
			window's selectedItem field.
			
			The wt and item parameters describe the window and item number of
			the What.  The current port will be the window associated with wt.
			
			If the What returns an error, the error will be displayed to the user.
		*)

		WhatClickProcPtr = function (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* The clickProc is called when the user clicks in the What.  

			The wt and item parameters describe the window and item number
			of the What.  The event is the original EventRecord.  The current port
			will be the window associated with wt.
			
			If the What returns an error, the error will be displayed to the user.
		*)

		WhatActivateProcPtr = function (wt: WindowType; item: integer; activate: Boolean): OSStatus;
		(* The activateProc is called when the window associated with the wt
			is activate or deactivated.  This event is broadcast to all Whats in
			the window.
			
			The wt and item parameters describe the window and item number
			of the What.  The activate Boolean is true if the window is being
			activate and false if it's being deactivated.  The current port
			will be the window associated with wt.
			
			Any error result is currently discarded.
		*)

		WhatCursorProcPtr = function (wt: WindowType; item: integer; mouseLocation: Point; cursorID: integer): OSStatus;
		(* The cursorProc is called when the cursor is inside the What.
			
			The wt and item parameters describe the window and item number
			of the What.  The mouseLocation is the current position of the mouse,
			in window local coordinates.  The cursorID is taken from the
			cursorID field of the WhatClassRecord.  The current port will be
			the window associated with wt.
			
			Any error result is currently discarded.
		*)

		WhatClassRecord =
			record
				whatTypeID: OSType;
				initProc: WhatInitProcPtr;
				openProc: WhatOpenProcPtr;
				flushProc: WhatFlushProcPtr;
				closeProc: WhatCloseProcPtr;
				keyProc: WhatKeyProcPtr;
				clickProc: WhatClickProcPtr;
				activateProc: WhatActivateProcPtr;
				cursorProc: WhatCursorProcPtr;
				cursorID: integer;
			end;
		WhatClassRecordPtr = ^WhatClassRecord;

	(* This array holds the WhatClassRecords for each of the classes
		in the system.
	*)
	var
		gWhatClasses: array[1..kMaxClass] of WhatClassRecord;

	(* ***** What Record ***** *)

	(* The WhatRecord holds the instance data for a What. This includes
		a pointer to the What's class, the IC key and flags associated with the
		What (taken from the What Template), and a bunch of instance
		data this is available for the What's private use.
		
		Instance data is under the control of the What Class.  The
		class can use the whatData, whatSpareData, whatFileSpec,
		and whatModified fields in any way it pleases.  The normal
		behaviour is for the What to allocate its instance data in its
		openProc and put a pointer to it in the whatData field.  The
		closeProc is then responsible for deallocate it.
		
		Obviously this instance data handling mechanism is exceedingly
		kludgy.  I have a good plan to solve this problem, but it will
		take a lot of effort to rework all the Whats, so I'm putting it
		off for the moment.
	*)
	
	type
		WhatRecord =
			record
				whatKey: Str31;
				whatClass: WhatClassRecordPtr;
				whatFlags: longint;
				whatData: Ptr;
				whatSpareData: Ptr;
				whatFileSpec: FSSpec;
				whatModified: Boolean;
			end;
		WhatRecordPtr = ^WhatRecord;

	(* ***** What Flags ***** *)

	(* The What Flags are initialised from the What Template and are
		used as indicators to the What Class as to how to handle this
		instance.
		
		The flags are roughly divided into two categories: flags common
		to all Whats, and flags dependent on the What Class.  Flags common
		to all whats are allocated from the highest bit downwards,
		class dependent flags are allocated from the bottom upwards.
		The exact dividing line is still undecided (-:
		
		The only common flag is the locked bit defined below. The What
		is responsible for setting this bit if its IC preference is locked,
		and for obeying the semantics of the lock.  This module exports utility
		routines (defined below) to assist with this.
		
		The class dependent flags are normally setup by the What Template
		and are defined in the What's module.
	*)
		
	const
		kWhatFlagsLockedBit = 31;
		kWhatFlagsLockedMask = $80000000;

	(* ***** Window List ***** *)

	(* The gWindows list holds the the data for all of the windows that
		IC manages. The fields include
		
		window				Holds a pointer to the window's system dialog structure.

		items					An array of pointers to the What's associated with the
									dialog items in this window. Note that any of these
									pointers may be nil.  This radically cuts down on the
									memory usage for windows that contain a lot of
									items that don't require Whats (like static text items)
									but it does complicate programming a bit.

		sendTabsToWhats	This Boolean controls whether Tab key presses should
									be sent to the selectedItem, or whether they should
									be managed by ICWindows. If the window contains
									any Text Whats, sendTabsToWhats is set to false
									because we currently have no way of arbitrating
									the keyboard focus between Text Whats and other
									Whats. If the window contains no text items, then
									other Whats may be eligable to receive Tab key presses.
									For example, the File Map What uses Tabs to cycle
									alphabetically through the list of mappings.

		selectedItem		This either holds the current keyboard focus,
									or -1 if there is none. If the window contains
									no Text Whats, a what can set this to the item
									it wishes to receive keyboard events. If the window
									contains Text Whats, this field is managed by
									ICWindows.

		position				Holds the window position to save. Not used by Whats.
	*)
	
	var
		gWindows: array[WindowType] of
			record
				window: DialogPtr;
				items: array[1..kMaxItemsPerWindow] of WhatRecordPtr;
				sendTabsToWhats:Boolean;
				selectedItem: integer;
				position: Point;
			end;

	(* ***** What Utilities ***** *)

	procedure WhatDirtyDocument;
		(* Whats can call this routine when they want to dirty the
			IC document.  It's basically a reexport of DirtyDocument
			so that the Whats can get at it without importing ICDocument.
		*)
	
	function WhatGetDocumentInstance: ICInstance;
		(* Whats can call this routine when they need an ICInstance
			to talk to the IC API.  It's basically a reexport of GetDocumentInstance
			so that the Whats can get at it without importing ICDocument.
		*)

	procedure WhatProcessAttributes (wt: WindowType; item: integer; attr: ICAttr);
		(* This routine processes the attributes return by the IC API
			and sets the kWhatFlagsLockedBit for the what if it needs to be.
		*)
	
	function WhatIsLocked (wt: WindowType; item: integer): Boolean;
		(* This routine returns true if the What is locked.  This can either
			be because the kWhatFlagsLockedBit is set, or because the
			document as a whole is locked.
		*)
	
	procedure WhatLockedAlert (wt: WindowType; item: integer);
		(* Whats should call this routine when uses attempt to modified
			a locked What. This routine keeps track of these attempts,
			and eventually notifies the user that they're not getting anywhere.
		*)

	function MapWindowPtrToWindowType (wp: WindowPtr): WindowType;
		(* This function searches gWindows and returns the WindowType
			that corresponds to the given WindowPtr.  What's call this routine
			when in system callbacks where they have access to the WindowPtr,
			but not to global variables.
		*)

	function MapWindowTypeToResourceID (wt: WindowType): integer;
		(* This (trivial) function returns the resource ID associated with
			the supplied window type. There is rarely a need for Whats
			to call this routine.
		*)
	
	function GetSelectedTextItem (wt: WindowType): integer;
		(* This function returns the item number of the Text What
			that's the current keyboard focus. It differs from
			gWindows[wt].selectedItem because it won't return
			a non-Text item.  This routine is rarely called by Whats except
			the Text What.
		*)

	(* ***** Initialisation ***** *)
	
	function InitICWindowGlobals : OSStatus;
		(* This routine initialises the module, including setting up
			the gWhatClasses array.  It also initialises each of
			of the What Classes.
		*)

implementation

	uses
		Dialogs, 

		ICMiscSubs, 
		ICResources,

		ICAboutWhats, ICSoundPopupWhats, ICRGBColourWhats, 
			ICGreyBoxWhats, ICCheckboxWhats, ICTextWhats, ICPopupWhats, 
				ICFontWhats, ICFSSpecWhats, ICFileMapWhat, ICHelperWhat, ICButtonWhat, 

		ICDocument;

	(* ***** What Utilities ***** *)

	procedure WhatDirtyDocument;
		(* See comment in interface part. *)
	begin
		DirtyDocument;
	end; (* WhatDirtyDocument *)

	function WhatGetDocumentInstance: ICInstance;
		(* See comment in interface part. *)
	begin
		WhatGetDocumentInstance := GetDocumentInstance;
	end; (* WhatGetDocumentInstance *)

	procedure WhatProcessAttributes (wt: WindowType; item: integer; attr: ICAttr);
		(* See comment in interface part. *)
	begin
		if (attr <> ICattr_no_change) & btst(attr, ICattr_locked_bit) then begin
			bset(gWindows[wt].items[item]^.whatFlags, kWhatFlagsLockedBit);
		end; (* if *)
	end; (* WhatProcessAttributes *)

	function WhatIsLocked (wt: WindowType; item: integer): Boolean;
		(* See comment in interface part. *)
	begin
		WhatIsLocked := btst(gWindows[wt].items[item]^.whatFlags, kWhatFlagsLockedBit) | IsDocumentLocked;
	end; (* WhatIsLocked *)

	(* These variables keep track of which locked window/item the user last
		attempted to modify, so that if the keep doing it we will eventually
		put up an alert telling them what's going on.
	*)
	
	var
		gLastLockedAlertWindow: WindowType;
		gLastLockedAlertItem: integer;

	procedure WhatLockedAlert (wt: WindowType; item: integer);
		(* See comment in interface part. *)
		var
			junk: integer;
	begin
		if (wt <> gLastLockedAlertWindow) or (item <> gLastLockedAlertItem) then begin
			ResetAlertStage;
			gLastLockedAlertWindow := wt;
			gLastLockedAlertItem := item;
		end; (* if *)
		InitCursor;
		junk := StopAlert(rLockedAlert, nil);
	end; (* WhatLockedAlert *)

	function MapWindowPtrToWindowType (wp: WindowPtr): WindowType;
		(* See comment in interface part. *)
		var
			i: WindowType;
	begin
		MapWindowPtrToWindowType := WT_None;
		if wp <> nil then begin
			for i := WT_None to WT_Last do begin
				if gWindows[i].window = wp then begin
					MapWindowPtrToWindowType := i;
					leave;	(* It's not my fault! -- Quinn *)
				end; (* if *)
			end; (* for *)
		end; (* if *)
	end; (* MapWindowPtrToWindowType *)

	function MapWindowTypeToResourceID (wt: WindowType): integer;
		(* See comment in interface part. *)
	begin
		if wt = WT_About then begin
			MapWindowTypeToResourceID := rAboutDialog;
		end else begin
			MapWindowTypeToResourceID := rMainDialog + ord(wt) - ord(WT_Main);
		end; (* if *)
	end; (* MapWindowTypeToResourceID *)

	function GetSelectedTextItem (wt: WindowType): integer;
		(* See comment in interface part. *)
		var
			selectedItem: integer;
	begin
		selectedItem := gWindows[wt].selectedItem;
		if (selectedItem > 0) & 
				(gWindows[wt].items[selectedItem] <> nil) & 
					(gWindows[wt].items[selectedItem]^.whatClass^.whatTypeID = 'TEXT') then begin
			GetSelectedTextItem := selectedItem;
		end else begin
			GetSelectedTextItem := -1;
		end; (* if *)
	end; (* GetSelectedTextItem *)

	(* ***** Initialisation ***** *)

	procedure InitWhatClasses;
		(* This routine initialises the gWhatClasses array.
			Structured constants? Who needs them!?!
		*)
		
		procedure W (
				whatClassIndex: integer;
				xtyp: OSType;
				xinit: WhatInitProcPtr;
				xopen : WhatOpenProcPtr;
				xflush : WhatFlushProcPtr;
				xclose : WhatCloseProcPtr;
				xkey : WhatKeyProcPtr;
				xclick : WhatClickProcPtr;
				xactivate : WhatActivateProcPtr;
				xcursor : WhatCursorProcPtr;
				xcursorid: integer);
		begin
			gWhatClasses[whatClassIndex].whatTypeID := xtyp;
			gWhatClasses[whatClassIndex].initProc := xinit;
			gWhatClasses[whatClassIndex].openProc := xopen;
			gWhatClasses[whatClassIndex].flushProc := xflush;
			gWhatClasses[whatClassIndex].closeProc := xclose;
			gWhatClasses[whatClassIndex].keyProc := xkey;
			gWhatClasses[whatClassIndex].clickProc := xclick;
			gWhatClasses[whatClassIndex].activateProc := xactivate;
			gWhatClasses[whatClassIndex].cursorProc := xcursor;
			gWhatClasses[whatClassIndex].cursorID := xcursorid;
		end; (* W *)
		
	begin
		W(1, 'NULL', nil, nil, nil, nil, nil, nil, nil, nil, 0);
 		W(2, 'TEXT', @WhatInitText, @WhatOpenText, @WhatFlushText, @WhatCloseText, @WhatKeyText, @WhatClickText, nil, nil, iBeamCursor);
		W(3, 'SPOP', nil, @WhatOpenPopup, @WhatFlushPopup, @WhatClosePopup, nil, @WhatClickPopup, nil, nil, 0);
		W(4, 'FFSP', nil, @WhatOpenFSSpec, @WhatFlushFSSpec, nil, nil, @WhatClickFSSpec, nil, nil, 0);
		W(5, 'FPOP', @WhatInitFont, @WhatOpenFont, @WhatFlushFont, nil, nil, @WhatClickFont, nil, nil, 0);
		W(6, 'FMAP', @WhatInitFileMap, @WhatOpenFileMap, @WhatFlushFileMap, @WhatCloseFileMap, @WhatKeyFileMap, @WhatClickFileMap, @WhatActivateFileMap, @WhatCursorFileMap, plusCursor);
		W(7, 'FBUT', nil, nil, nil, nil, nil, @WhatClickFileMap, nil, nil, 0);
		W(8, 'HMAP', @WhatInitHelper, @WhatOpenHelper, @WhatFlushHelper, @WhatCloseHelper, @WhatKeyHelper, @WhatClickHelper, @WhatActivateHelper, @WhatCursorHelper, plusCursor);
		W(9, 'HBUT', nil, nil, nil, nil, nil, @WhatClickHelper, nil, nil, 0);
		W(10, 'BUTN', @WhatInitButton, @WhatOpenButton, nil, nil, nil, @WhatClickButton, nil, nil, 0);
		W(11, 'FSIZ', nil, nil, nil, nil, nil, @WhatClickFontSize, nil, nil, 0);
		W(12, 'CHKB', nil, @WhatOpenCheckbox, @WhatFlushCheckbox, nil, nil, @WhatClickCheckbox, nil, nil, 0);
		W(13, 'GREY', @WhatInitGreyBox, @WhatOpenGreyBox, nil, nil, nil, nil, nil, nil, 0);
		W(14, 'RGBC', @WhatInitRGBColour, @WhatOpenRGBColour, @WhatFlushRGBColour, @WhatCloseRGBColour, nil, @WhatClickRGBColour, nil, nil, 0);
		W(15, 'SNDP', nil, @WhatOpenSoundPopup, @WhatFlushSoundPopup, nil, nil, @WhatClickSoundPopup, nil, nil, 0);
		W(16, 'BOUT', @WhatInitAbout, @WhatOpenAbout, nil, nil, nil, @WhatClickAbout, nil, nil, 0);
	end; (* InitWhatClasses *)

	function InitICWindowGlobals : OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
			firstErr : OSStatus;
			wt: WindowType;
			whatClassIndex : integer;
	begin
		InitWhatClasses;
		firstErr := noErr;
		for whatClassIndex := 1 to kMaxClass do begin
			if gWhatClasses[whatClassIndex].initProc <> nil then begin
				err := gWhatClasses[whatClassIndex].initProc(0);
				if firstErr = noErr then begin
					firstErr := err;
				end; (* if *)
			end; (* if *)
		end; (* for *)
		for wt := WT_None to WT_Last do begin
			gWindows[wt].window := nil;
		end; (* for *)
		gLastLockedAlertWindow := WT_None;
		gLastLockedAlertItem := -1;
		InitICWindowGlobals := firstErr;
	end; (* InitICWindowGlobals *)

end. (* ICWindowGlobals *)
